fix pr44

Alan Modra amodra@bigpond.net.au
Wed Oct 13 02:27:00 GMT 2004


On Wed, Oct 13, 2004 at 09:05:31AM +0930, Alan Modra wrote:
> still fails for me though.  A little debugging shows that the
> reloc in .text for i has been reduced to one against .data, but
> .data is just a local symbol rather than being the section symbol.

I suppose there isn't anything particularly wrong with a local symbol
that has the same name as a section.  This brings to light a deficiency
in ld's cross-reference checking, which is that only global and section
symbols are currently checked.

	PR 44
	* ldcref.c (check_section_sym_xref): Delete.
	(check_local_sym_xref): New function.
	(check_nocrossrefs): Adjust call.

Index: ld/ldcref.c
===================================================================
RCS file: /cvs/src/src/ld/ldcref.c,v
retrieving revision 1.9
diff -u -p -r1.9 ldcref.c
--- ld/ldcref.c	28 Jun 2003 05:28:54 -0000	1.9
+++ ld/ldcref.c	13 Oct 2004 02:21:56 -0000
@@ -69,7 +69,7 @@ struct cref_hash_table {
 /* Forward declarations.  */
 
 static void output_one_cref (FILE *, struct cref_hash_entry *);
-static void check_section_sym_xref (lang_input_statement_type *);
+static void check_local_sym_xref (lang_input_statement_type *);
 static bfd_boolean check_nocrossref (struct cref_hash_entry *, void *);
 static void check_refs (const char *, asection *, bfd *,
 			struct lang_nocrossrefs *);
@@ -136,7 +136,7 @@ cref_hash_newfunc (struct bfd_hash_entry
 }
 
 /* Add a symbol to the cref hash table.  This is called for every
-   symbol that is seen during the link.  */
+   global symbol that is seen during the link.  */
 
 void
 add_cref (const char *name,
@@ -330,39 +330,69 @@ check_nocrossrefs (void)
 
   cref_hash_traverse (&cref_table, check_nocrossref, NULL);
 
-  lang_for_each_file (check_section_sym_xref);
+  lang_for_each_file (check_local_sym_xref);
 }
 
 /* Checks for prohibited cross references to section symbols.  */
 
 static void
-check_section_sym_xref (lang_input_statement_type *statement)
+check_local_sym_xref (lang_input_statement_type *statement)
 {
   bfd *abfd;
-  asection *sec;
+  lang_input_statement_type *li;
+  asymbol **asymbols, **syms;
 
   abfd = statement->the_bfd;
   if (abfd == NULL)
     return;
 
-  for (sec = abfd->sections; sec != NULL; sec = sec->next)
+  li = abfd->usrdata;
+  if (li != NULL && li->asymbols != NULL)
+    asymbols = li->asymbols;
+  else
     {
-      asection *outsec;
+      long symsize;
+      long symbol_count;
 
-      outsec = sec->output_section;
-      if (outsec != NULL)
+      symsize = bfd_get_symtab_upper_bound (abfd);
+      if (symsize < 0)
+	einfo (_("%B%F: could not read symbols; %E\n"), abfd);
+      asymbols = xmalloc (symsize);
+      symbol_count = bfd_canonicalize_symtab (abfd, asymbols);
+      if (symbol_count < 0)
+	einfo (_("%B%F: could not read symbols: %E\n"), abfd);
+      if (li != NULL)
+	{
+	  li->asymbols = asymbols;
+	  li->symbol_count = symbol_count;
+	}
+    }
+
+  for (syms = asymbols; *syms; ++syms)
+    {
+      asymbol *sym = *syms;
+      if (sym->flags & (BSF_GLOBAL | BSF_WARNING | BSF_INDIRECT | BSF_FILE))
+	continue;
+      if ((sym->flags & (BSF_LOCAL | BSF_SECTION_SYM)) != 0
+	  && sym->section->output_section != NULL)
 	{
-	  const char *outsecname;
+	  const char *outsecname, *symname;
 	  struct lang_nocrossrefs *ncrs;
 	  struct lang_nocrossref *ncr;
 
-	  outsecname = outsec->name;
+	  outsecname = sym->section->output_section->name;
+	  symname = NULL;
+	  if ((sym->flags & BSF_SECTION_SYM) == 0)
+	    symname = sym->name;
 	  for (ncrs = nocrossref_list; ncrs != NULL; ncrs = ncrs->next)
 	    for (ncr = ncrs->list; ncr != NULL; ncr = ncr->next)
 	      if (strcmp (ncr->name, outsecname) == 0)
-		check_refs (NULL, sec, abfd, ncrs);
+		check_refs (symname, sym->section, abfd, ncrs);
 	}
     }
+
+  if (li == NULL)
+    free (asymbols);
 }
 
 /* Check one symbol to see if it is a prohibited cross reference.  */
@@ -330,39 +330,69 @@ check_nocrossrefs (void)
 }
 
-/* Checks for prohibited cross references to section symbols.  */
+/* Check for prohibited cross references to local and section symbols.  */
 
 static void

-- 
Alan Modra
IBM OzLabs - Linux Technology Centre



More information about the Binutils mailing list